#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2017-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import torch
import numpy as np
from typing import Optional, Text

from .base import Application

from pyannote.core import Segment, Timeline, Annotation

from pyannote.database import get_protocol
from pyannote.database import get_annotated
from pyannote.database import Subset
from pyannote.database import get_unique_identifier
from pyannote.database import FileFinder
from pyannote.database.protocol import SpeakerDiarizationProtocol
from pyannote.database.protocol import SpeakerVerificationProtocol

import scipy.optimize
from scipy.cluster.hierarchy import fcluster
from pyannote.core.utils.hierarchy import linkage

from pyannote.core.utils.distance import pdist
from pyannote.core.utils.distance import cdist
from pyannote.audio.features.precomputed import Precomputed

from pyannote.metrics.binary_classification import det_curve
from pyannote.metrics.diarization import DiarizationPurityCoverageFMeasure

from pyannote.audio.features import Pretrained
from pyannote.audio.features.utils import get_audio_duration


class SpeakerEmbedding(Application):
    @property
    def config_default_module(self):
        return "pyannote.audio.embedding.approaches"

    def validation_criterion(self, protocol_name, **kwargs):
        protocol = get_protocol(protocol_name)
        if isinstance(protocol, SpeakerVerificationProtocol):
            return "equal_error_rate"
        elif isinstance(protocol, SpeakerDiarizationProtocol):
            return "diarization_fscore"

    def validate_init(self, protocol_name: Text, subset: Subset = "development"):

        protocol = get_protocol(protocol_name)

        if isinstance(
            protocol, (SpeakerVerificationProtocol, SpeakerDiarizationProtocol)
        ):
            return

        msg = (
            "Only SpeakerVerification or SpeakerDiarization tasks are"
            'supported in "validation" mode.'
        )
        raise ValueError(msg)

    def validate_epoch(self, epoch, validation_data, protocol=None, **kwargs):

        _protocol = get_protocol(protocol)

        if isinstance(_protocol, SpeakerVerificationProtocol):
            return self._validate_epoch_verification(
                epoch, validation_data, protocol=protocol, **kwargs
            )

        elif isinstance(_protocol, SpeakerDiarizationProtocol):
            return self._validate_epoch_diarization(
                epoch, validation_data, protocol=protocol, **kwargs
            )

        else:
            msg = (
                "Only SpeakerVerification or SpeakerDiarization tasks are"
                'supported in "validation" mode.'
            )
            raise ValueError(msg)

    @staticmethod
    def get_hash(file):
        hashable = []
        for f in file.files():
            hashable.append((f["uri"], tuple(f["try_with"])))
        return hash(tuple(sorted(hashable)))

    @staticmethod
    def get_embedding(file, pretrained):
        emb = []
        for f in file.files():
            if isinstance(f["try_with"], Segment):
                segments = [f["try_with"]]
            else:
                segments = f["try_with"]
            for segment in segments:
                # Crop preferentially with "center", but use "loose"
                # if "center" produces empty clip
                for mode in ["center", "loose"]:
                    x = pretrained.crop(f, segment, mode=mode)
                    if len(x) > 0:
                        break
                emb.append(x)

        return np.mean(np.vstack(emb), axis=0, keepdims=True)

    def _validate_epoch_verification(
        self,
        epoch,
        validation_data,
        protocol=None,
        subset: Subset = "development",
        device: Optional[torch.device] = None,
        batch_size: int = 32,
        n_jobs: int = 1,
        duration: float = None,
        step: float = 0.25,
        metric: str = None,
        **kwargs,
    ):

        # initialize embedding extraction
        pretrained = Pretrained(
            validate_dir=self.validate_dir_,
            epoch=epoch,
            duration=duration,
            step=step,
            batch_size=batch_size,
            device=device,
        )

        preprocessors = self.preprocessors_
        if "audio" not in preprocessors:
            preprocessors["audio"] = FileFinder()
        if "duration" not in preprocessors:
            preprocessors["duration"] = get_audio_duration
        _protocol = get_protocol(protocol, preprocessors=preprocessors)

        y_true, y_pred, cache = [], [], {}

        for trial in getattr(_protocol, f"{subset}_trial")():

            # compute embedding for file1
            file1 = trial["file1"]
            hash1 = self.get_hash(file1)
            if hash1 in cache:
                emb1 = cache[hash1]
            else:
                emb1 = self.get_embedding(file1, pretrained)
                cache[hash1] = emb1

            # compute embedding for file2
            file2 = trial["file2"]
            hash2 = self.get_hash(file2)
            if hash2 in cache:
                emb2 = cache[hash2]
            else:
                emb2 = self.get_embedding(file2, pretrained)
                cache[hash2] = emb2

            # compare embeddings
            distance = cdist(emb1, emb2, metric=metric)[0, 0]
            y_pred.append(distance)

            y_true.append(trial["reference"])
        _, _, _, eer = det_curve(np.array(y_true), np.array(y_pred), distances=True)

        return {"metric": "equal_error_rate", "minimize": True, "value": float(eer)}

    def _validate_epoch_diarization(
        self,
        epoch,
        validation_data,
        protocol=None,
        subset: Subset = "development",
        device: Optional[torch.device] = None,
        batch_size: int = 32,
        n_jobs: int = 1,
        duration: float = None,
        step: float = 0.25,
        metric: str = None,
        **kwargs,
    ):

        # initialize embedding extraction
        pretrained = Pretrained(
            validate_dir=self.validate_dir_,
            epoch=epoch,
            duration=duration,
            step=step,
            batch_size=batch_size,
            device=device,
        )

        preprocessors = self.preprocessors_
        if "audio" not in preprocessors:
            preprocessors["audio"] = FileFinder()
        if "duration" not in preprocessors:
            preprocessors["duration"] = get_audio_duration
        _protocol = get_protocol(protocol, preprocessors=preprocessors)

        Z, t = dict(), dict()
        min_d, max_d = np.inf, -np.inf

        for current_file in getattr(_protocol, subset)():

            uri = get_unique_identifier(current_file)
            uem = get_annotated(current_file)
            reference = current_file["annotation"]

            X_, t_ = [], []
            embedding = pretrained(current_file)
            for i, (turn, _) in enumerate(reference.itertracks()):

                # extract embedding for current speech turn
                x_ = embedding.crop(turn, mode="center")
                if len(x_) < 1:
                    x_ = embedding.crop(turn, mode="loose")
                if len(x_) < 1:
                    msg = f"No embedding for {turn} in {uri:s}."
                    raise ValueError(msg)

                # each speech turn is represented by its average embedding
                X_.append(np.mean(x_, axis=0))
                t_.append(turn)

            X_ = np.array(X_)
            # apply hierarchical agglomerative clustering
            # all the way up to just one cluster (ie complete dendrogram)
            D = pdist(X_, metric=metric)
            min_d = min(np.min(D), min_d)
            max_d = max(np.max(D), max_d)

            Z[uri] = linkage(X_, method="pool", metric=metric)
            t[uri] = np.array(t_)

        def fun(threshold):

            _metric = DiarizationPurityCoverageFMeasure(weighted=False)

            for current_file in getattr(_protocol, subset)():

                uri = get_unique_identifier(current_file)
                uem = get_annotated(current_file)
                reference = current_file["annotation"]

                clusters = fcluster(Z[uri], threshold, criterion="distance")

                hypothesis = Annotation(uri=uri)
                for (start_time, end_time), cluster in zip(t[uri], clusters):
                    hypothesis[Segment(start_time, end_time)] = cluster

                _ = _metric(reference, hypothesis, uem=uem)

            return 1.0 - abs(_metric)

        res = scipy.optimize.minimize_scalar(
            fun, bounds=(0.0, 1.0), method="bounded", options={"maxiter": 10}
        )

        threshold = res.x.item()

        return {
            "metric": "diarization_fscore",
            "minimize": False,
            "value": float(1.0 - res.fun),
        }



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Tasks
#####

This module provides a `Task` class meant to specify machine learning tasks
(e.g. classification or regression).

This may be used to infer parts of the network architecture and the associated
loss function automatically.

Example
-------
>>> voice_activity_detection = Task(type=TaskType.MULTI_CLASS_CLASSIFICATION,
...                                 output=TaskOutput.SEQUENCE)
"""

from enum import Enum
from typing import Union
from typing import NamedTuple
from typing import Callable


class TaskType(Enum):
    """Type of machine learning task

    Attributes
    ----------
    MULTI_CLASS_CLASSIFICATION
        multi-class classification
    MULTI_LABEL_CLASSIFICATION
        multi-label classification
    REGRESSION
        regression
    REPRESENTATION_LEARNING
        representation learning
    """

    MULTI_CLASS_CLASSIFICATION = 0
    MULTI_LABEL_CLASSIFICATION = 1
    REGRESSION = 2
    REPRESENTATION_LEARNING = 3


class TaskOutput(Enum):
    """Expected output

    Attributes
    ----------
    SEQUENCE
        A sequence of vector is expected.
    VECTOR
        A single vector is expected.
    """

    SEQUENCE = 0
    VECTOR = 1


class Task(NamedTuple):
    type: TaskType
    output: TaskOutput

    @classmethod
    def from_str(cls, representation: str):
        task_output, task_type = representation.split(" ", 1)

        if task_output == "frame-wise":
            task_output = TaskOutput.SEQUENCE

        elif task_output == "chunk-wise":
            task_output = TaskOutput.VECTOR

        else:
            msg = f'"{task_output}" task output is not supported.'
            raise NotImplementedError(msg)

        if task_type == "multi-class classification":
            task_type = TaskType.MULTI_CLASS_CLASSIFICATION

        elif task_type == "multi-label classification":
            task_type = TaskType.MULTI_LABEL_CLASSIFICATION

        elif task_type == "regression":
            task_type = TaskType.REGRESSION

        elif task_type == "representation learning":
            task_type = TaskType.REPRESENTATION_LEARNING

        else:
            msg = f'"{task_type}" task type is not supported.'
            raise NotImplementedError(msg)

        return cls(type=task_type, output=task_output)

    def __str__(self) -> str:
        """String representation"""

        if self.returns_sequence:
            name = "frame-wise"

        elif self.returns_vector:
            name = "chunk-wise"

        else:
            msg = (
                "string representation (__str__) is not implemented "
                "for this task output."
            )
            raise NotImplementedError(msg)

        if self.is_multiclass_classification:
            name = f"{name} multi-class classification"

        elif self.is_multilabel_classification:
            name = f"{name} multi-label classification"

        elif self.is_regression:
            name = f"{name} regression"

        elif self.is_representation_learning:
            name = f"{name} representation learning"

        else:
            msg = (
                "string representation (__str__) is not implemented "
                "for this type of task."
            )
            raise NotImplementedError(msg)

        return name

    @property
    def returns_sequence(self) -> bool:
        """Is the output expected to be a sequence?

        Returns
        -------
        `bool`
            `True` if the task output is a sequence, `False` otherwise.
        """
        return self.output == TaskOutput.SEQUENCE

    @property
    def returns_vector(self) -> bool:
        """Is the output expected to be a single vector?

        Returns
        -------
        `bool`
            `True` if the task output is a single vector, `False` otherwise.
        """
        return self.output == TaskOutput.VECTOR

    @property
    def is_multiclass_classification(self) -> bool:
        """Is it multi-class classification?

        Returns
        -------
        `bool`
            `True` if the task is multi-class classification
        """
        return self.type == TaskType.MULTI_CLASS_CLASSIFICATION

    @property
    def is_multilabel_classification(self) -> bool:
        """Is it multi-label classification?

        Returns
        -------
        `bool`
            `True` if the task is multi-label classification
        """
        return self.type == TaskType.MULTI_LABEL_CLASSIFICATION

    @property
    def is_regression(self) -> bool:
        """Is it regression?

        Returns
        -------
        `bool`
            `True` if the task is regression
        """
        return self.type == TaskType.REGRESSION

    @property
    def is_representation_learning(self) -> bool:
        """Is it representation learning?

        Returns
        -------
        `bool`
            `True` if the task is representation learning
        """
        return self.type == TaskType.REPRESENTATION_LEARNING

    @property
    def default_activation(self):
        """Default final activation

        Returns
        -------
        `torch.nn.LogSoftmax(dim=-1)` for multi-class classification
        `torch.nn.Sigmoid()` for multi-label classification
        `torch.nn.Identity()` for regression

        Raises
        ------
        NotImplementedError
            If the default activation cannot be guessed.
        """

        import torch.nn

        if self.is_multiclass_classification:
            return torch.nn.LogSoftmax(dim=-1)

        elif self.is_multilabel_classification:
            return torch.nn.Sigmoid()

        elif self.is_regression:
            return torch.nn.Identity()

        else:
            msg = f"Unknown default activation for {self} task."
            raise NotImplementedError(msg)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional
import numpy as np
import torch
from pyannote.core.utils.distance import to_condensed
from scipy.spatial.distance import squareform
from .base import RepresentationLearning


class TripletLoss(RepresentationLearning):
    """Triplet loss

    TODO explain

    Parameters
    ----------
    duration : float, optional
        Chunks duration, in seconds. Defaults to 1.
    min_duration : float, optional
        When provided, use chunks of random duration between `min_duration` and
        `duration` for training. Defaults to using fixed duration chunks.
    per_turn : int, optional
        Number of chunks per speech turn. Defaults to 1.
        If per_turn is greater than one, embeddings of the same speech turn
        are averaged before comparison. The intuition is that it might help
        learn embeddings meant to be averaged/summed.
    per_label : `int`, optional
        Number of sequences per speaker in each batch. Defaults to 1.
    per_fold : `int`, optional
        Number of different speakers per batch. Defaults to 32.
    per_epoch : `float`, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    label_min_duration : `float`, optional
        Remove speakers with less than that many seconds of speech.
        Defaults to 0 (i.e. keep them all).
    metric : {'euclidean', 'cosine', 'angular'}, optional
        Defaults to 'cosine'.
    margin: float, optional
        Margin multiplicative factor. Defaults to 0.2.
    clamp : {'positive', 'sigmoid', 'softmargin'}, optional
        Defaults to 'positive'.
    sampling : {'all', 'hard', 'negative', 'easy'}, optional
        Triplet sampling strategy. Defaults to 'all' (i.e. use all possible
        triplets). 'hard' sampling use both hardest positive and negative for
        each anchor. 'negative' sampling use hardest negative for each
        (anchor, positive) pairs. 'easy' sampling only use easy triplets (i.e.
        those for which d(anchor, positive) < d(anchor, negative)).

    Notes
    -----
    delta = d(anchor, positive) - d(anchor, negative)

    * with 'positive' clamping:
        loss = max(0, delta + margin x D)
    * with 'sigmoid' clamping:
        loss = sigmoid(10 * delta)

    where d(., .) varies in range [0, D] (e.g. D=2 for euclidean distance).

    Reference
    ---------
    TODO
    """

    def __init__(
        self,
        duration: float = 1.0,
        min_duration: float = None,
        per_turn: int = 1,
        per_label: int = 1,
        per_fold: int = 32,
        per_epoch: Optional[float] = None,
        label_min_duration: float = 0.0,
        # FIXME create a Literal type for metric
        # FIXME maybe in pyannote.core.utils.distance
        metric: str = "cosine",
        # FIXME homogeneize the meaning of margin parameter
        # FIXME it has a different meaning in ArcFace, right?
        margin: float = 0.2,
        # FIXME create a Literal type for clamp
        clamp="positive",
        # FIXME create a Literal type for sampling
        sampling="all",
    ):

        super().__init__(
            duration=duration,
            min_duration=min_duration,
            per_turn=per_turn,
            per_label=per_label,
            per_fold=per_fold,
            per_epoch=per_epoch,
            label_min_duration=label_min_duration,
        )

        self.metric = metric
        self.margin = margin
        # FIXME see above
        self.margin_ = self.margin * self.max_distance

        if clamp not in {"positive", "sigmoid", "softmargin"}:
            msg = "'clamp' must be one of {'positive', 'sigmoid', 'softmargin'}."
            raise ValueError(msg)
        self.clamp = clamp

        if sampling not in {"all", "hard", "negative", "easy"}:
            msg = "'sampling' must be one of {'all', 'hard', 'negative', 'easy'}."
            raise ValueError(msg)
        self.sampling = sampling

    def batch_easy(self, y, distances):
        """Build easy triplets"""

        anchors, positives, negatives = [], [], []

        distances = squareform(self.to_numpy(distances))

        for anchor, y_anchor in enumerate(y):
            for positive, y_positive in enumerate(y):

                # if same embedding or different labels, skip
                if (anchor == positive) or (y_anchor != y_positive):
                    continue

                d = distances[anchor, positive]

                for negative, y_negative in enumerate(y):

                    if y_negative == y_anchor:
                        continue

                    if d > distances[anchor, negative]:
                        continue

                    anchors.append(anchor)
                    positives.append(positive)
                    negatives.append(negative)

        return anchors, positives, negatives

    def batch_hard(self, y, distances):
        """Build triplet with both hardest positive and hardest negative

        Parameters
        ----------
        y : list
            Sequence labels.
        distances : (n * (n-1) / 2,) torch.Tensor
            Condensed pairwise distance matrix

        Returns
        -------
        anchors, positives, negatives : list of int
            Triplets indices.
        """

        anchors, positives, negatives = [], [], []

        distances = squareform(self.to_numpy(distances))
        y = np.array(y)

        for anchor, y_anchor in enumerate(y):

            d = distances[anchor]

            # hardest positive
            pos = np.where(y == y_anchor)[0]
            pos = [p for p in pos if p != anchor]
            positive = int(pos[np.argmax(d[pos])])

            # hardest negative
            neg = np.where(y != y_anchor)[0]
            negative = int(neg[np.argmin(d[neg])])

            anchors.append(anchor)
            positives.append(positive)
            negatives.append(negative)

        return anchors, positives, negatives

    def batch_negative(self, y, distances):
        """Build triplet with hardest negative

        Parameters
        ----------
        y : list
            Sequence labels.
        distances : (n * (n-1) / 2,) torch.Tensor
            Condensed pairwise distance matrix

        Returns
        -------
        anchors, positives, negatives : list of int
            Triplets indices.
        """

        anchors, positives, negatives = [], [], []

        distances = squareform(self.to_numpy(distances))
        y = np.array(y)

        for anchor, y_anchor in enumerate(y):

            # hardest negative
            d = distances[anchor]
            neg = np.where(y != y_anchor)[0]
            negative = int(neg[np.argmin(d[neg])])

            for positive in np.where(y == y_anchor)[0]:
                if positive == anchor:
                    continue

                anchors.append(anchor)
                positives.append(positive)
                negatives.append(negative)

        return anchors, positives, negatives

    def batch_all(self, y, distances):
        """Build all possible triplet

        Parameters
        ----------
        y : list
            Sequence labels.
        distances : (n * (n-1) / 2,) torch.Tensor
            Condensed pairwise distance matrix

        Returns
        -------
        anchors, positives, negatives : list of int
            Triplets indices.
        """

        anchors, positives, negatives = [], [], []

        for anchor, y_anchor in enumerate(y):
            for positive, y_positive in enumerate(y):

                # if same embedding or different labels, skip
                if (anchor == positive) or (y_anchor != y_positive):
                    continue

                for negative, y_negative in enumerate(y):

                    if y_negative == y_anchor:
                        continue

                    anchors.append(anchor)
                    positives.append(positive)
                    negatives.append(negative)

        return anchors, positives, negatives

    def triplet_loss(self, distances, anchors, positives, negatives):
        """Compute triplet loss

        Parameters
        ----------
        distances : torch.Tensor
            Condensed matrix of pairwise distances.
        anchors, positives, negatives : list of int
            Triplets indices.

        Returns
        -------
        loss : torch.Tensor
            Triplet loss.
        """

        # estimate total number of embeddings from pdist shape
        n = int(0.5 * (1 + np.sqrt(1 + 8 * len(distances))))

        # convert indices from squared matrix
        # to condensed matrix referential
        pos = to_condensed(n, anchors, positives)
        neg = to_condensed(n, anchors, negatives)

        # compute raw triplet loss (no margin, no clamping)
        # the lower, the better
        delta = distances[pos] - distances[neg]

        # clamp triplet loss
        if self.clamp == "positive":
            loss = torch.clamp(delta + self.margin_, min=0)

        elif self.clamp == "softmargin":
            loss = torch.log1p(torch.exp(delta))

        elif self.clamp == "sigmoid":
            # TODO. tune this "10" hyperparameter
            # TODO. log-sigmoid
            loss = torch.sigmoid(10 * (delta + self.margin_))

        return loss

    def batch_loss(self, batch):
        """Compute loss for current `batch`

        Parameters
        ----------
        batch : `dict`
            ['X'] (`numpy.ndarray`)
            ['y'] (`numpy.ndarray`)

        Returns
        -------
        batch_loss : `dict`
            ['loss'] (`torch.Tensor`) : Triplet loss
        """

        fX, y = self.embed(batch)
        distances = self.pdist(fX)

        # sample triplets
        triplets = getattr(self, "batch_{0}".format(self.sampling))
        anchors, positives, negatives = triplets(y, distances)

        # compute loss for each triplet
        losses = self.triplet_loss(distances, anchors, positives, negatives)

        loss = torch.mean(losses)

        # average over all triplets
        return {"loss": loss, "loss_triplet": loss}



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import numpy as np
from typing import Optional

from pyannote.core import Annotation
from pyannote.core import Timeline
from pyannote.core.utils.numpy import one_hot_decoding
from pyannote.pipeline import Pipeline
from pyannote.audio.features import Precomputed
from pyannote.pipeline.blocks.clustering import HierarchicalAgglomerativeClustering
from pyannote.pipeline.blocks.clustering import AffinityPropagationClustering
from .utils import assert_string_labels

from pyannote.audio.features.wrapper import Wrapper, Wrappable


class SpeechTurnClustering(Pipeline):
    """Speech turn clustering

    Parameters
    ----------
    embedding : Wrappable, optional
        Describes how raw speaker embeddings should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
        Defaults to "@emb" that indicates that protocol files provide
        the scores in the "emb" key.
    metric : {'euclidean', 'cosine', 'angular'}, optional
        Metric used for comparing embeddings. Defaults to 'cosine'.
    method : {'pool', 'affinity_propagation'}
        Set method used for clustering. "pool" stands for agglomerative
        hierarchical clustering with embedding pooling. "affinity_propagation"
        is for clustering based on affinity propagation. Defaults to "pool".
    window_wise : `bool`, optional
        Set `window_wise` to True to apply clustering on embedding extracted
        using the built-in sliding window. Defaults to apply clustering at
        speech turn level (one average embedding per speech turn).
    """

    def __init__(
        self,
        embedding: Wrappable = None,
        metric: Optional[str] = "cosine",
        method: Optional[str] = "pool",
        window_wise: Optional[bool] = False,
    ):
        super().__init__()

        if embedding is None:
            embedding = "@emb"
        self.embedding = embedding
        self._embedding = Wrapper(self.embedding)

        self.metric = metric
        self.method = method

        if self.method == "affinity_propagation":
            self.clustering = AffinityPropagationClustering(metric=self.metric)

            # sklearn documentation: Preferences for each point - points with
            # larger values of preferences are more likely to be chosen as
            # exemplars. The number of exemplars, ie of clusters, is influenced by
            # the input preferences value. If the preferences are not passed as
            # arguments, they will be set to the median of the input similarities.

            # NOTE one could set the preference value of each speech turn
            # according to their duration. longer speech turns are expected to
            # have more accurate embeddings, therefore should be prefered for
            # exemplars

        else:
            self.clustering = HierarchicalAgglomerativeClustering(
                method=self.method, metric=self.metric, use_threshold=True
            )

        self.window_wise = window_wise

    def _window_level(self, current_file: dict, speech_regions: Timeline) -> Annotation:
        """Apply clustering at window level

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        speech_regions : `Timeline`
            Speech regions.

        Returns
        -------
        hypothesis : `pyannote.core.Annotation`
            Clustering result.
        """

        # load embeddings
        embedding = self._embedding(current_file)
        window = embedding.sliding_window

        # extract and stack embeddings of speech regions
        X = np.vstack(
            [
                embedding.crop(segment, mode="center", fixed=segment.duration)
                for segment in speech_regions
            ]
        )

        # apply clustering
        y_pred = self.clustering(X)

        # reconstruct
        y = np.zeros(len(embedding), dtype=np.int8)

        # n = total number of "speech" embeddings
        # s_pred = current position in y_pred
        s_pred, n = 0, len(y_pred)

        for segment in speech_regions:

            # get indices of current speech segment
            ((s, e),) = window.crop(
                segment, mode="center", fixed=segment.duration, return_ranges=True
            )

            # hack for the very last segment that might overflow by 1
            e_pred = min(s_pred + e - s, n - 1)
            e = s + (e_pred - s_pred)

            # assign y_pred to the corresponding speech regions
            y[s:e] = y_pred[s_pred:e_pred]

            # increment current position in y_red
            s_pred += e - s

        # reconstruct hypothesis
        return one_hot_decoding(y, window)

    def _turn_level(self, current_file: dict, speech_turns: Annotation) -> Annotation:
        """Apply clustering at speech turn level

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        speech_turns : `Annotation`
            Speech turns. Should only contain `str` labels.

        Returns
        -------
        hypothesis : `pyannote.core.Annotation`
            Clustering result.
        """

        assert_string_labels(speech_turns, "speech_turns")

        embedding = self._embedding(current_file)

        labels = speech_turns.labels()
        X, clustered_labels, skipped_labels = [], [], []
        for l, label in enumerate(labels):

            timeline = speech_turns.label_timeline(label, copy=False)

            # be more and more permissive until we have
            # at least one embedding for current speech turn
            for mode in ["strict", "center", "loose"]:
                x = embedding.crop(timeline, mode=mode)
                if len(x) > 0:
                    break

            # skip labels so small we don't have any embedding for it
            if len(x) < 1:
                skipped_labels.append(label)
                continue

            clustered_labels.append(label)
            X.append(np.mean(x, axis=0))

        # apply clustering of label embeddings
        clusters = self.clustering(np.vstack(X))

        # map each clustered label to its cluster (between 1 and N_CLUSTERS)
        mapping = {label: k for label, k in zip(clustered_labels, clusters)}

        # map each skipped label to its own cluster
        # (between -1 and -N_SKIPPED_LABELS)
        for l, label in enumerate(skipped_labels):
            mapping[label] = -(l + 1)

        # do the actual mapping
        return speech_turns.rename_labels(mapping=mapping)

    def __call__(
        self, current_file: dict, speech_turns: Optional[Annotation] = None
    ) -> Annotation:
        """Apply speech turn clustering

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        speech_turns : `Annotation`, optional
            Speech turns. Should only contain `str` labels.
            Defaults to `current_file['speech_turns']`.

        Returns
        -------
        speech_turns : `pyannote.core.Annotation`
            Clustered speech turns (or windows in case `window_wise` is True)
        """

        if speech_turns is None:
            speech_turns = current_file["speech_turns"]

        if self.window_wise:
            return self._window_level(
                current_file, speech_turns.get_timeline().support()
            )

        return self._turn_level(current_file, speech_turns)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016-2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import versioneer

from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="pyannote.audio",
    namespace_packages=["pyannote"],
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "pyannote-audio=pyannote.audio.applications.pyannote_audio:main",
            "pyannote-speech-feature=pyannote.audio.applications.feature_extraction:main",
        ],
        "prodigy_recipes": [
            "pyannote.sad.manual = pyannote.audio.interactive.recipes.sad:sad_manual",
            "pyannote.dia.binary = pyannote.audio.interactive.recipes.dia:dia_binary",
            "pyannote.dia.manual = pyannote.audio.interactive.recipes.dia:dia_manual",
        ],
    },
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Neural building blocks for speaker diarization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Hervé Bredin",
    author_email="bredin@limsi.fr",
    url="https://github.com/pyannote/pyannote-audio",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import numpy as np
from .base_labeling import BaseLabeling
from pyannote.audio.features import Pretrained
from collections import Counter
from sklearn.metrics import confusion_matrix

import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt


def plot_confusion_matrix(
    y_true, y_pred, classes, normalize=False, title=None, cmap=plt.cm.Blues
):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if not title:
        if normalize:
            title = "Normalized confusion matrix"
        else:
            title = "Confusion matrix, without normalization"

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    if normalize:
        cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print("Confusion matrix, without normalization")

    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
    ax.figure.colorbar(im, ax=ax)
    # We want to show all ticks...
    ax.set(
        xticks=np.arange(cm.shape[1]),
        yticks=np.arange(cm.shape[0]),
        # ... and label them with the respective list entries
        xticklabels=classes,
        yticklabels=classes,
        title=title,
        ylabel="True label",
        xlabel="Predicted label",
    )

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = ".1f" if normalize else "d"
    thresh = cm.max() / 2.0
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            if cm[i, j] == 0.0:
                continue
            ax.text(
                j,
                i,
                format(cm[i, j], fmt),
                ha="center",
                va="center",
                color="white" if cm[i, j] > thresh else "black",
                size=6,
            )
    fig.tight_layout()
    return ax


class DomainClassification(BaseLabeling):
    def validate_epoch(
        self,
        epoch,
        validation_data,
        device=None,
        batch_size=32,
        n_jobs=1,
        duration=None,
        step=0.25,
        **kwargs
    ):

        pretrained = Pretrained(
            validate_dir=self.validate_dir_,
            epoch=epoch,
            duration=duration,
            step=step,
            batch_size=batch_size,
            device=device,
        )

        domain = self.task_.domain
        domains = pretrained.classes

        y_true_file, y_pred_file = [], []

        for current_file in validation_data:

            y_pred = pretrained(current_file).data.argmax(axis=1)
            y_pred_file.append(Counter(y_pred).most_common(1)[0][0])

            y_true = domains.index(current_file[domain])
            y_true_file.append(y_true)

        accuracy = np.mean(np.array(y_true_file) == np.array(y_pred_file))

        return {"metric": "accuracy", "minimize": False, "value": float(accuracy)}



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import warnings
from functools import partial

from pathlib import Path
import shutil
from typing import Text
from typing import Dict
import yaml

import collections
from pyannote.core.utils.helper import get_class_by_name
from pyannote.database import FileFinder
from pyannote.audio.features.utils import get_audio_duration
from pyannote.audio.train.task import Task


def merge_cfg(pretrained_cfg, cfg):
    for k, v in cfg.items():

        # case where the user purposedly set a section value to "null"
        # this might happen when fine-tuning a pretrained model
        if v is None:
            _ = pretrained_cfg.pop(k, None)

        # if v is a dictionary, go deeper and merge recursively
        elif isinstance(v, collections.abc.Mapping):
            pretrained_cfg[k] = merge_cfg(pretrained_cfg.get(k, {}), v)

        # in any other case, override pretrained_cfg[k] by cfg[k]
        else:
            pretrained_cfg[k] = v

    return pretrained_cfg


def load_config(
    config_yml: Path,
    training: bool = False,
    config_default_module: Text = None,
    pretrained_config_yml: Path = None,
) -> Dict:
    """

    Returns
    -------
    config : Dict
        ['preprocessors']
        ['learning_rate']
        ['scheduler']
        ['get_optimizer']
        ['callbacks']
        ['feature_extraction']
        ['task']
        ['get_model_from_specs']
        ['model_resolution']
        ['model_alignment']
    """

    # load pretrained model configuration
    pretrained_cfg = dict()
    if pretrained_config_yml is not None:
        with open(pretrained_config_yml, "r") as fp:
            pretrained_cfg = yaml.load(fp, Loader=yaml.SafeLoader)

    # load configuration or complain it's missing
    cfg = dict()
    if config_yml.exists():
        with open(config_yml, "r") as fp:
            cfg = yaml.load(fp, Loader=yaml.SafeLoader)

        # backup user-provided config because it will be updated
        if pretrained_config_yml is not None:
            shutil.copy(config_yml, config_yml.parent / "backup+config.yml")

    elif pretrained_config_yml is None:
        msg = f"{config_yml} configuration file is missing."
        raise FileNotFoundError(msg)

    # override pretrained model config with user-provided config
    cfg = merge_cfg(pretrained_cfg, cfg)

    # save (updated) config to disk
    if pretrained_config_yml is not None:
        with open(config_yml, "w") as fp:
            yaml.dump(cfg, fp, default_flow_style=False)

    # preprocessors
    preprocessors = dict()

    for key, preprocessor in cfg.get("preprocessors", {}).items():
        # preprocessors:
        #    key:
        #       name: package.module.ClassName
        #       params:
        #          param1: value1
        #          param2: value2
        if isinstance(preprocessor, dict):
            Klass = get_class_by_name(preprocessor["name"])
            preprocessors[key] = Klass(**preprocessor.get("params", {}))
            continue

        try:
            # preprocessors:
            #    key: /path/to/database.yml
            preprocessors[key] = FileFinder(database_yml=preprocessor)

        except FileNotFoundError as e:
            # preprocessors:
            #    key: /path/to/{uri}.wav
            preprocessors[key] = preprocessor

    cfg["preprocessors"] = preprocessors

    # scheduler
    SCHEDULER_DEFAULT = {
        "name": "DavisKingScheduler",
        "params": {"learning_rate": "auto"},
    }
    scheduler_cfg = cfg.get("scheduler", SCHEDULER_DEFAULT)
    Scheduler = get_class_by_name(
        scheduler_cfg["name"], default_module_name="pyannote.audio.train.schedulers"
    )
    scheduler_params = scheduler_cfg.get("params", {})

    cfg["learning_rate"] = scheduler_params.pop("learning_rate", "auto")
    cfg["scheduler"] = Scheduler(**scheduler_params)

    # optimizer
    OPTIMIZER_DEFAULT = {
        "name": "SGD",
        "params": {
            "momentum": 0.9,
            "dampening": 0,
            "weight_decay": 0,
            "nesterov": True,
        },
    }
    optimizer_cfg = cfg.get("optimizer", OPTIMIZER_DEFAULT)
    try:
        Optimizer = get_class_by_name(
            optimizer_cfg["name"], default_module_name="torch.optim"
        )
        optimizer_params = optimizer_cfg.get("params", {})
        cfg["get_optimizer"] = partial(Optimizer, **optimizer_params)

    # do not raise an error here as it is possible that the optimizer is
    # not really needed (e.g. in pipeline training)
    except ModuleNotFoundError as e:
        warnings.warn(e.args[0])

    # data augmentation should only be active when training a model
    if training and "data_augmentation" in cfg:
        DataAugmentation = get_class_by_name(
            cfg["data_augmentation"]["name"],
            default_module_name="pyannote.audio.augmentation",
        )
        augmentation = DataAugmentation(**cfg["data_augmentation"].get("params", {}))
    else:
        augmentation = None

    # custom callbacks
    callbacks = []
    for callback_config in cfg.get("callbacks", {}):
        Callback = get_class_by_name(callback_config["name"])
        callback = Callback(**callback_config.get("params", {}))
        callbacks.append(callback)
    cfg["callbacks"] = callbacks

    # feature extraction
    FEATURE_DEFAULT = {"name": "RawAudio", "params": {"sample_rate": 16000}}
    feature_cfg = cfg.get("feature_extraction", FEATURE_DEFAULT)
    FeatureExtraction = get_class_by_name(
        feature_cfg["name"], default_module_name="pyannote.audio.features"
    )
    feature_params = feature_cfg.get("params", {})
    cfg["feature_extraction"] = FeatureExtraction(
        **feature_params, augmentation=augmentation
    )

    # task
    if config_default_module is None:
        config_default_module = "pyannote.audio.labeling.tasks"

    try:
        TaskClass = get_class_by_name(
            cfg["task"]["name"], default_module_name=config_default_module
        )
    except AttributeError:
        TaskClass = get_class_by_name(
            cfg["task"]["name"],
            default_module_name="pyannote.audio.embedding.approaches",
        )

    cfg["task"] = TaskClass(**cfg["task"].get("params", {}))

    # architecture
    Architecture = get_class_by_name(
        cfg["architecture"]["name"], default_module_name="pyannote.audio.models"
    )
    params = cfg["architecture"].get("params", {})

    cfg["get_model_from_specs"] = partial(Architecture, **params)
    task = cfg["task"].task
    cfg["model_resolution"] = Architecture.get_resolution(task, **params)
    cfg["model_alignment"] = Architecture.get_alignment(task, **params)

    return cfg


def load_specs(specs_yml: Path) -> Dict:
    """

    Returns
    -------
    specs : Dict
        ['task']
        [and others]
    """

    with open(specs_yml, "r") as fp:
        specifications = yaml.load(fp, Loader=yaml.SafeLoader)
    specifications["task"] = Task.from_str(specifications["task"])
    return specifications


def load_params(params_yml: Path) -> Dict:

    with open(params_yml, "r") as fp:
        params = yaml.load(fp, Loader=yaml.SafeLoader)

    return params



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Speech activity detection"""

from typing import Optional
from typing import Text

import numpy as np
import torch
import torch.nn as nn
from .base import LabelingTask
from .base import LabelingTaskGenerator
from pyannote.audio.train.task import Task, TaskType, TaskOutput
from ..gradient_reversal import GradientReversal
from pyannote.audio.models.models import RNN
from pyannote.audio.features.wrapper import Wrappable
from pyannote.database import Protocol
from pyannote.database import Subset
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import Alignment


class SpeechActivityDetectionGenerator(LabelingTaskGenerator):
    """Batch generator for training speech activity detection

    Parameters
    ----------
    task : Task
        Task
    feature_extraction : Wrappable
        Describes how features should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
    protocol : Protocol
    subset : {'train', 'development', 'test'}, optional
        Protocol and subset.
    resolution : `pyannote.core.SlidingWindow`, optional
        Override `feature_extraction.sliding_window`. This is useful for
        models that include the feature extraction step (e.g. SincNet) and
        therefore output a lower sample rate than that of the input.
        Defaults to `feature_extraction.sliding_window`
    alignment : {'center', 'loose', 'strict'}, optional
        Which mode to use when cropping labels. This is useful for models that
        include the feature extraction step (e.g. SincNet) and therefore use a
        different cropping mode. Defaults to 'center'.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    mask : str, optional
        When provided, protocol files are expected to contain a key named after
        this `mask` variable and providing a `SlidingWindowFeature` instance.
        Generated batches will contain an additional "mask" key (on top of
        existing "X" and "y" keys) computed as an excerpt of `current_file[mask]`
        time-aligned with "y". Defaults to not add any "mask" key.

    """

    def __init__(
        self,
        task: Task,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        resolution: Optional[Resolution] = None,
        alignment: Optional[Alignment] = None,
        duration: float = 2.0,
        batch_size: int = 32,
        per_epoch: float = None,
        mask: Text = None,
    ):

        super().__init__(
            task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=duration,
            batch_size=batch_size,
            per_epoch=per_epoch,
            exhaustive=False,
            mask=mask,
            local_labels=True,
        )

    def postprocess_y(self, Y: np.ndarray) -> np.ndarray:
        """Generate labels for speech activity detection

        Parameters
        ----------
        Y : (n_samples, n_speakers) numpy.ndarray
            Discretized annotation returned by
            `pyannote.core.utils.numpy.one_hot_encoding`.

        Returns
        -------
        y : (n_samples, 1) numpy.ndarray

        See also
        --------
        `pyannote.core.utils.numpy.one_hot_encoding`
        """

        # number of speakers for each frame
        speaker_count = np.sum(Y, axis=1, keepdims=True)

        # mark speech regions as such
        return np.int64(speaker_count > 0)

    @property
    def specifications(self):
        specs = {
            "task": self.task,
            "X": {"dimension": self.feature_extraction.dimension},
            "y": {"classes": ["non_speech", "speech"]},
        }

        for key, classes in self.file_labels_.items():

            # TODO. add an option to handle this list
            # TODO. especially useful for domain-adversarial stuff
            if key in ["duration", "audio", "uri"]:
                continue
            specs[key] = {"classes": classes}

        return specs


class SpeechActivityDetection(LabelingTask):
    """Train speech activity (and overlap) detection

    Parameters
    ----------
    duration : float, optional
        Duration of sub-sequences. Defaults to 3.2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Total audio duration per epoch, in days.
        Defaults to one day (1).
    """

    def get_batch_generator(
        self,
        feature_extraction,
        protocol,
        subset: Subset = "train",
        resolution=None,
        alignment=None,
    ):
        """
        resolution : `pyannote.core.SlidingWindow`, optional
            Override `feature_extraction.sliding_window`. This is useful for
            models that include the feature extraction step (e.g. SincNet) and
            therefore output a lower sample rate than that of the input.
        alignment : {'center', 'loose', 'strict'}, optional
            Which mode to use when cropping labels. This is useful for models
            that include the feature extraction step (e.g. SincNet) and
            therefore use a different cropping mode. Defaults to 'center'.
        """
        return SpeechActivityDetectionGenerator(
            self.task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=self.duration,
            per_epoch=self.per_epoch,
            batch_size=self.batch_size,
        )


class DomainAwareSpeechActivityDetection(SpeechActivityDetection):
    """Domain-aware speech activity detection

    Trains speech activity detection and domain classification jointly.

    Parameters
    ----------
    domain : `str`, optional
        Batch key to use as domain. Defaults to 'domain'.
        Could be 'database' or 'uri' for instance.
    attachment : `int`, optional
        Intermediate level where to attach the domain classifier.
        Defaults to -1. Passed to `return_intermediate` in models supporting it.
    rnn: `dict`, optional
        Parameters of the RNN used in the domain classifier.
        See `pyannote.audio.models.models.RNN` for details.
    domain_loss : `str`, optional
        Loss function to use. Defaults to 'NLLLoss'.
    """

    DOMAIN_PT = "{train_dir}/weights/{epoch:04d}.domain.pt"

    def __init__(
        self, domain="domain", attachment=-1, rnn=None, domain_loss="NLLLoss", **kwargs
    ):
        super().__init__(**kwargs)
        self.domain = domain
        self.attachment = attachment

        if rnn is None:
            rnn = dict()
        self.rnn = rnn

        self.domain_loss = domain_loss
        if self.domain_loss == "NLLLoss":
            # Default value
            self.domain_loss_ = nn.NLLLoss()
            self.activation_ = nn.LogSoftmax(dim=1)

        elif self.domain_loss == "MSELoss":
            self.domain_loss_ = nn.MSELoss()
            self.activation_ = nn.Sigmoid()

        else:
            msg = f"{domain_loss} has not been implemented yet."
            raise NotImplementedError(msg)

    def more_parameters(self):
        """Initialize trainable trainer parameters

        Yields
        ------
        parameter : nn.Parameter
            Trainable trainer parameters
        """

        domain_classifier_rnn = RNN(
            n_features=self.model.intermediate_dimension(self.attachment), **self.rnn
        )

        n_classes = len(self.specifications[self.domain]["classes"])
        domain_classifier_linear = nn.Linear(
            domain_classifier_rnn.dimension, n_classes, bias=True
        ).to(self.device)

        self.domain_classifier_ = nn.Sequential(
            domain_classifier_rnn, domain_classifier_linear
        ).to(self.device)

        # TODO: check if we really need to do this .to(self.device) twice

        return self.domain_classifier_.parameters()

    def load_more(self, model_pt=None) -> bool:
        """Load classifier from disk"""

        if model_pt is None:
            domain_pt = self.DOMAIN_PT.format(
                train_dir=self.train_dir_, epoch=self.epoch_
            )
        else:
            domain_pt = model_pt.with_suffix(".domain.pt")

        domain_classifier_state = torch.load(
            domain_pt, map_location=lambda storage, loc: storage
        )
        self.domain_classifier_.load_state_dict(domain_classifier_state)

        # FIXME add support for different domains
        return True

    def save_more(self):
        """Save domain classifier to disk"""

        domain_pt = self.DOMAIN_PT.format(train_dir=self.train_dir_, epoch=self.epoch_)
        torch.save(self.domain_classifier_.state_dict(), domain_pt)

    def batch_loss(self, batch):
        """Compute loss for current `batch`

        Parameters
        ----------
        batch : `dict`
            ['X'] (`numpy.ndarray`)
            ['y'] (`numpy.ndarray`)

        Returns
        -------
        batch_loss : `dict`
            ['loss'] (`torch.Tensor`) : Loss
        """

        # forward pass
        X = torch.tensor(batch["X"], dtype=torch.float32, device=self.device_)
        fX, intermediate = self.model_(X, return_intermediate=self.attachment)

        # speech activity detection
        fX = fX.view((-1, self.n_classes_))
        target = (
            torch.tensor(batch["y"], dtype=torch.int64, device=self.device_)
            .contiguous()
            .view((-1,))
        )

        weight = self.weight
        if weight is not None:
            weight = weight.to(device=self.device_)
        loss = self.loss_func_(fX, target, weight=weight)

        # domain classification
        domain_target = torch.tensor(
            batch[self.domain], dtype=torch.int64, device=self.device_
        )

        domain_scores = self.activation_(self.domain_classifier_(intermediate))

        domain_loss = self.domain_loss_(domain_scores, domain_target)

        return {
            "loss": loss + domain_loss,
            "loss_domain": domain_loss,
            "loss_task": loss,
        }


class DomainAdversarialSpeechActivityDetection(DomainAwareSpeechActivityDetection):
    """Domain Adversarial speech activity detection

    Parameters
    ----------
    domain : `str`, optional
        Batch key to use as domain. Defaults to 'domain'.
        Could be 'database' or 'uri' for instance.
    attachment : `int`, optional
        Intermediate level where to attach the domain classifier.
        Defaults to -1. Passed to `return_intermediate` in models supporting it.
    alpha : `float`, optional
        Coefficient multiplied with the domain loss
    """

    def __init__(self, domain="domain", attachment=-1, alpha=1.0, **kwargs):
        super().__init__(domain=domain, attachment=attachment, **kwargs)
        self.alpha = alpha
        self.gradient_reversal_ = GradientReversal()

    def batch_loss(self, batch):
        """Compute loss for current `batch`

        Parameters
        ----------
        batch : `dict`
            ['X'] (`numpy.ndarray`)
            ['y'] (`numpy.ndarray`)

        Returns
        -------
        batch_loss : `dict`
            ['loss'] (`torch.Tensor`) : Loss
        """
        # forward pass
        X = torch.tensor(batch["X"], dtype=torch.float32, device=self.device_)

        fX, intermediate = self.model_(X, return_intermediate=self.attachment)

        # speech activity detection
        fX = fX.view((-1, self.n_classes_))

        target = (
            torch.tensor(batch["y"], dtype=torch.int64, device=self.device_)
            .contiguous()
            .view((-1,))
        )

        weight = self.weight
        if weight is not None:
            weight = weight.to(device=self.device_)

        loss = self.loss_func_(fX, target, weight=weight)

        # domain classification
        domain_target = torch.tensor(
            batch[self.domain], dtype=torch.int64, device=self.device_
        )

        domain_scores = self.activation_(
            self.domain_classifier_(self.gradient_reversal_(intermediate))
        )

        if self.domain_loss == "MSELoss":
            # One hot encode domain_target for Mean Squared Error Loss
            nb_domains = domain_scores.shape[1]
            identity_mat = torch.sparse.torch.eye(nb_domains, device=self.device_)
            domain_target = identity_mat.index_select(dim=0, index=domain_target)

        domain_loss = self.domain_loss_(domain_scores, domain_target)

        return {
            "loss": loss + self.alpha * domain_loss,
            "loss_domain": domain_loss,
            "loss_task": loss,
        }



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2017-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from typing import Optional
import numpy as np

from pyannote.pipeline import Pipeline
from pyannote.pipeline.blocks.classification import ClosestAssignment
from pyannote.core import Annotation
from .utils import assert_int_labels
from .utils import assert_string_labels
from ..features import Precomputed

from pyannote.audio.features.wrapper import Wrapper, Wrappable


class SpeechTurnClosestAssignment(Pipeline):
    """Assign speech turn to closest cluster

    Parameters
    ----------
    embedding : Wrappable, optional
        Describes how raw speaker embeddings should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
        Defaults to "@emb" that indicates that protocol files provide
        the scores in the "emb" key.
    metric : {'euclidean', 'cosine', 'angular'}, optional
        Metric used for comparing embeddings. Defaults to 'cosine'.
    """

    def __init__(self, embedding: Wrappable = None, metric: Optional[str] = "cosine"):
        super().__init__()

        if embedding is None:
            embedding = "@emb"
        self.embedding = embedding
        self._embedding = Wrapper(self.embedding)

        self.metric = metric

        self.closest_assignment = ClosestAssignment(metric=self.metric)

    def __call__(
        self, current_file: dict, speech_turns: Annotation, targets: Annotation
    ) -> Annotation:
        """Assign each speech turn to closest target (if close enough)

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        speech_turns : `Annotation`
            Speech turns. Should only contain `int` labels.
        targets : `Annotation`
            Targets. Should only contain `str` labels.

        Returns
        -------
        assigned : `Annotation`
            Assigned speech turns.
        """

        assert_string_labels(targets, "targets")
        assert_int_labels(speech_turns, "speech_turns")

        embedding = self._embedding(current_file)

        # gather targets embedding
        labels = targets.labels()
        X_targets, targets_labels = [], []
        for l, label in enumerate(labels):

            timeline = targets.label_timeline(label, copy=False)

            # be more and more permissive until we have
            # at least one embedding for current speech turn
            for mode in ["center", "loose"]:
                x = embedding.crop(timeline, mode=mode)
                if len(x) > 0:
                    break

            # skip labels so small we don't have any embedding for it
            if len(x) < 1:
                continue

            targets_labels.append(label)
            X_targets.append(np.mean(x, axis=0))

        # gather speech turns embedding
        labels = speech_turns.labels()
        X, assigned_labels, skipped_labels = [], [], []
        for l, label in enumerate(labels):

            timeline = speech_turns.label_timeline(label, copy=False)

            # be more and more permissive until we have
            # at least one embedding for current speech turn
            for mode in ["center", "loose"]:
                x = embedding.crop(timeline, mode=mode)
                if len(x) > 0:
                    break

            # skip labels so small we don't have any embedding for it
            if len(x) < 1:
                skipped_labels.append(label)
                continue

            assigned_labels.append(label)
            X.append(np.mean(x, axis=0))

        # assign speech turns to closest class
        assignments = self.closest_assignment(np.vstack(X_targets), np.vstack(X))
        mapping = {
            label: targets_labels[k]
            for label, k in zip(assigned_labels, assignments)
            if not k < 0
        }
        return speech_turns.rename_labels(mapping=mapping)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional
from typing import Text

import numpy as np
from pyannote.database import get_annotated
from pyannote.core import Segment
from pyannote.core import SlidingWindow
from pyannote.core import Timeline

from pyannote.core.utils.random import random_segment
from pyannote.core.utils.random import random_subsegment

from pyannote.audio.features import RawAudio

from .base import LabelingTask
from .base import LabelingTaskGenerator
from pyannote.audio.train.task import Task, TaskType, TaskOutput

from pyannote.audio.features.wrapper import Wrappable
from pyannote.database import Protocol
from pyannote.database import Subset
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import Alignment


normalize = lambda wav: wav / (np.sqrt(np.mean(wav ** 2)) + 1e-8)


class OverlapDetectionGenerator(LabelingTaskGenerator):
    """Batch generator for training overlap detection

    Parameters
    ----------
    task : Task
        Task
    feature_extraction : Wrappable
        Describes how features should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
    protocol : Protocol
    subset : {'train', 'development', 'test'}, optional
        Protocol and subset.
    resolution : `pyannote.core.SlidingWindow`, optional
        Override `feature_extraction.sliding_window`. This is useful for
        models that include the feature extraction step (e.g. SincNet) and
        therefore output a lower sample rate than that of the input.
        Defaults to `feature_extraction.sliding_window`
    alignment : {'center', 'loose', 'strict'}, optional
        Which mode to use when cropping labels. This is useful for models that
        include the feature extraction step (e.g. SincNet) and therefore use a
        different cropping mode. Defaults to 'center'.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    snr_min, snr_max : float, optional
        Defines Signal-to-Overlap Ratio range in dB. Defaults to [0, 10].
    """

    def __init__(
        self,
        task: Task,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        resolution: Optional[Resolution] = None,
        alignment: Optional[Alignment] = None,
        duration: float = 2.0,
        batch_size: int = 32,
        per_epoch: float = None,
        snr_min: float = 0,
        snr_max: float = 10,
    ):

        self.snr_min = snr_min
        self.snr_max = snr_max
        self.raw_audio_ = RawAudio(sample_rate=feature_extraction.sample_rate)

        super().__init__(
            task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=duration,
            batch_size=batch_size,
            per_epoch=per_epoch,
            # TODO. be smart and find a way to use "local_labels"
            local_labels=False,
        )

    def overlap_samples(self):
        """Random overlap samples

        Returns
        -------
        samples : generator
            Generator that yields {'waveform': ..., 'y': ...} samples
            indefinitely.
        """

        uris = list(self.data_)
        durations = np.array([self.data_[uri]["duration"] for uri in uris])
        probabilities = durations / np.sum(durations)

        while True:

            # choose file at random with probability
            # proportional to its (annotated) duration
            uri = uris[np.random.choice(len(uris), p=probabilities)]

            datum = self.data_[uri]
            current_file = datum["current_file"]

            # choose one segment at random with probability
            # proportional to its duration
            segment = next(random_segment(datum["segments"], weighted=True))

            # choose random subsegment
            # duration = np.random.rand() * self.duration
            sequence = next(random_subsegment(segment, self.duration))

            # get corresponding waveform
            X = self.raw_audio_.crop(
                current_file, sequence, mode="center", fixed=self.duration
            )

            # get corresponding labels
            y = datum["y"].crop(sequence, mode=self.alignment, fixed=self.duration)

            yield {"waveform": normalize(X), "y": y}

    def sliding_samples(self):
        """Sliding window

        Returns
        -------
        samples : generator
            Generator that yields {'waveform': ..., 'y': ...} samples
            indefinitely.
        """

        uris = list(self.data_)
        durations = np.array([self.data_[uri]["duration"] for uri in uris])
        probabilities = durations / np.sum(durations)

        sliding_segments = SlidingWindow(duration=self.duration, step=self.duration)

        while True:

            # shuffle files
            np.random.shuffle(uris)

            # loop on shuffled files
            for uri in uris:

                datum = self.data_[uri]

                # make a copy of current file
                current_file = dict(datum["current_file"])

                # read waveform for the whole file
                waveform = self.raw_audio_(current_file)

                # randomly shift 'annotated' segments start time so that
                # we avoid generating exactly the same subsequence twice
                shifted_segments = [
                    Segment(s.start + np.random.random() * self.duration, s.end)
                    for s in get_annotated(current_file)
                ]
                # deal with corner case where a shifted segment would be empty
                shifted_segments = [s for s in shifted_segments if s]
                annotated = Timeline(segments=shifted_segments)

                samples = []
                for sequence in sliding_segments(annotated):

                    X = waveform.crop(sequence, mode="center", fixed=self.duration)

                    y = datum["y"].crop(
                        sequence, mode=self.alignment, fixed=self.duration
                    )

                    # FIXME -- this is ugly
                    sample = {
                        "waveform": normalize(X),
                        "y": y,
                        "database": current_file["database"],
                        "uri": current_file["uri"],
                        "audio": current_file["audio"],
                        "duration": current_file["duration"],
                    }

                    samples.append(sample)

                np.random.shuffle(samples)
                for sample in samples:
                    yield sample

    def samples(self):
        """Training sample generator"""

        sliding_samples = self.sliding_samples()
        overlap_samples = self.overlap_samples()

        while True:

            # get fixed duration random sequence
            original = next(sliding_samples)

            if np.random.rand() < 0.5:
                pass

            else:
                # get random overlapping sequence
                overlap = next(overlap_samples)

                # select SNR at random
                snr = (
                    self.snr_max - self.snr_min
                ) * np.random.random_sample() + self.snr_min
                alpha = np.exp(-np.log(10) * snr / 20)

                original["waveform"] += alpha * overlap["waveform"]
                original["y"] += overlap["y"]

            speaker_count = np.sum(original["y"], axis=1, keepdims=True)
            original["y"] = np.int64(speaker_count > 1)

            # run feature extraction
            original["X"] = self.feature_extraction.crop(
                original, Segment(0, self.duration), mode="center", fixed=self.duration
            )

            yield {"X": original["X"], "y": original["y"]}

    @property
    def specifications(self):
        return {
            "task": self.task,
            "X": {"dimension": self.feature_extraction.dimension},
            "y": {"classes": ["non_overlap", "overlap"]},
        }


class OverlapDetection(LabelingTask):
    """Train overlap detection

    Parameters
    ----------
    duration : float, optional
        Duration of sub-sequences. Defaults to 3.2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Total audio duration per epoch, in days.
        Defaults to one day (1).
    """

    def get_batch_generator(
        self,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        resolution: Optional[Resolution] = None,
        alignment: Optional[Resolution] = None,
    ) -> OverlapDetectionGenerator:
        """Get batch generator

        Parameters
        ----------
        feature_extraction : Wrappable
            Describes how features should be obtained.
            See pyannote.audio.features.wrapper.Wrapper documentation for details.
        protocol : Protocol
        subset : {'train', 'development', 'test'}, optional
            Protocol and subset.
        resolution : `pyannote.core.SlidingWindow`, optional
            Override `feature_extraction.sliding_window`. This is useful for
            models that include the feature extraction step (e.g. SincNet) and
            therefore output a lower sample rate than that of the input.
        alignment : {'center', 'loose', 'strict'}, optional
            Which mode to use when cropping labels. This is useful for models
            that include the feature extraction step (e.g. SincNet) and
            therefore use a different cropping mode. Defaults to 'center'.
        """
        return OverlapDetectionGenerator(
            self.task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=self.duration,
            per_epoch=self.per_epoch,
            batch_size=self.batch_size,
        )



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import numpy as np
import scipy.stats
from collections import deque
from .callback import Callback
from tqdm import tqdm
from scipy.signal import convolve

from typing import TYPE_CHECKING

AUTOLR_MIN = 0.000001
AUTOLR_MAX = 10
AUTOLR_EPOCHS = 8
AUTOLR_BETA = 0.98

MOMENTUM_MAX = 0.95
MOMENTUM_MIN = 0.85

if False:
    from .trainer import Trainer


def decreasing_probability(values: np.ndarray) -> float:
    """Compute probability that a sequence is decreasing

    Parameters
    ----------
    values : np.ndarray
        Sequence of values

    Returns
    -------
    probability : float
        Probability that sequence of values is decreasing

    Reference
    ---------
    Davis King. "Automatic Learning Rate Scheduling That Really Works".
    http://blog.dlib.net/2018/02/automatic-learning-rate-scheduling-that.html
    """
    n_steps = len(values)
    steps = np.arange(n_steps)

    A = np.vstack([steps, np.ones(n_steps)]).T
    loc, shift = np.linalg.lstsq(A, values, rcond=None)[0]

    values_ = loc * steps + shift
    sigma2 = np.sum((values - values_) ** 2) / (n_steps - 2)

    scale = np.sqrt(12 * sigma2 / (n_steps ** 3 - n_steps))
    return scipy.stats.norm.cdf(0.0, loc=loc, scale=scale)


def steps_without_decrease(values: np.ndarray, robust: bool = False) -> int:
    """Count number of steps without decrease

    Parameters
    ----------
    values : np.ndarray
        Sequence of values
    robust : bool
        Remove 10% highest values before counting steps.

    Returns
    -------
    n_steps : int
        Number of steps

    Reference
    ---------
    Davis King. "Automatic Learning Rate Scheduling That Really Works".
    http://blog.dlib.net/2018/02/automatic-learning-rate-scheduling-that.html
    """

    if robust:
        values = values[values < np.percentile(values, 90)]

    steps_without_decrease = 0
    n_steps = len(values)
    for i in reversed(range(n_steps - 2)):
        p = decreasing_probability(values[i:])
        if p < 0.51:
            steps_without_decrease = n_steps - i
    return steps_without_decrease


class BaseSchedulerCallback(Callback):
    """Base scheduler with support for AutoLR

    Reference
    ---------
    Leslie N. Smith. "Cyclical Learning Rates for Training Neural Networks"
    IEEE Winter Conference on Applications of Computer Vision (WACV, 2017).
    """

    def on_train_start(self, trainer: "Trainer") -> None:
        self.optimizer_ = trainer.optimizer
        if trainer.base_learning_rate_ == "auto":
            trainer.base_learning_rate_ = self.auto_lr(trainer)
        self.learning_rate = trainer.base_learning_rate_

    def on_epoch_start(self, trainer: "Trainer") -> None:
        """Log learning rate to tensorboard"""

        trainer.tensorboard_.add_scalar(
            f"train/lr", self.learning_rate, global_step=trainer.epoch_
        )

    @property
    def learning_rate(self) -> float:
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, learning_rate: float) -> None:
        for g in self.optimizer_.param_groups:
            g["lr"] = learning_rate
            g["momentum"] = MOMENTUM_MAX
        self._learning_rate = learning_rate

    @staticmethod
    def _choose_lr(lrs: np.ndarray, losses: np.ndarray) -> float:
        """Choose learning rate upper bound

        This is done by selecting the [0.1 x lr, lr] range that leads to the
        largest decrease in terms of loss value.

        Parameters
        ----------
        lrs : np.ndarray
            Sequence of learning rates
        losses : np.ndarray
            Corresponding sequence of loss values

        Returns
        -------
        max_lr : float
            Return learning rate upper bound
        """

        min_lr, max_lr = np.min(lrs), np.max(lrs)
        n_batches = len(lrs)

        # `factor` by which the learning rate is multiplied after every batch,
        # to get from `min_lr` to `max_lr` in `n_batches` step.
        factor = (max_lr / min_lr) ** (1 / n_batches)

        # K batches to increase the learning rate by one order of magnitude
        K = int(np.log(10) / np.log(factor))

        # loss improvement on each [0.1 x lr, lr] range
        improvement = losses[:-K] - losses[K:]

        # return half LR such that [0.1 x lr, lr] leads to the best improvement
        return 0.5 * lrs[K + np.argmax(improvement)]

    def auto_lr(self, trainer: "Trainer") -> float:
        """Find optimal learning rate automatically

        Parameters
        ----------
        trainer : 'Trainer'

        Returns
        -------
        learning_rate : float
            Optimal learning rate

        Reference
        ---------
        Leslie N. Smith. "Cyclical Learning Rates for Training Neural Networks"
        IEEE Winter Conference on Applications of Computer Vision (WACV, 2017).

            There is a simple way to estimate reasonable minimum and maximum
            boundary values with one training run of the network for a few
            epochs. It is a "LR range test"; run your model for several epochs
            while letting the learning rate increase linearly between low and
            high LR values. This test is enormously valuable whenever you are
            facing a new architecture or dataset. Next, plot the accuracy
            versus learning rate. Note the learning rate value when the
            accuracy starts to increase and when the accuracy slows, becomes
            ragged, or starts to fall. These two learning rates are good
            choices for bounds; that is, set base_lr to the first value and
            set max_lr to the latter value. Alternatively, one can use the rule
            of thumb that the optimum learning rate is usually within a factor
            of two of the largest one that converges and set base_lr to 1/3 or
            1/4 of max_lr. [...] Whenever one is starting with a new
            architecture or dataset, a single LR range test provides both a
            good LR value and a good range. Then one should compare runs with a
            fixed LR versus CLR with this range. Whichever wins can be used
            with confidence for the rest of one’s experiments.
        """
        # save states to disk
        trainer.save_state()

        # initialize optimizer with a low learning rate
        self.learning_rate = AUTOLR_MIN

        # `factor` by which the learning rate is multiplied after every batch,
        # to get from `min_lr` to `max_lr` in `n_batches` step.
        n_batches: int = AUTOLR_EPOCHS * trainer.batches_per_epoch
        factor = (AUTOLR_MAX / AUTOLR_MIN) ** (1 / n_batches)

        # progress bar
        pbar = tqdm(
            desc="AutoLR",
            total=n_batches,
            leave=False,
            ncols=80,
            unit="batch",
            position=1,
        )

        loss_moving_avg = 0.0
        losses, losses_smoothened, lrs = [], [], []

        # loop on n_batches batches
        for i in range(n_batches):

            batch = trainer.get_new_batch()
            loss = trainer.batch_loss(batch)
            loss["loss"].backward()
            trainer.optimizer_.step()
            trainer.optimizer_.zero_grad()

            l = loss["loss"].item()
            if np.isnan(l):
                break

            losses.append(l)
            lrs.append(self.learning_rate)

            loss_moving_avg = (
                AUTOLR_BETA * loss_moving_avg + (1 - AUTOLR_BETA) * losses[-1]
            )
            losses_smoothened.append(loss_moving_avg / (1 - AUTOLR_BETA ** (i + 1)))

            # update progress bar
            pbar.update(1)
            pbar.set_postfix(
                ordered_dict={"loss": losses_smoothened[-1], "lr": self.learning_rate}
            )

            # increase learning rate
            self.learning_rate = factor * self.learning_rate

            # stop AutoLR early when loss starts to explode
            if i > 1 and losses_smoothened[-1] > 100 * np.nanmin(losses_smoothened):
                break

        # reload model using its initial state
        trainer.load_state()

        # choose learning rate based on loss = f(learning_rate) curve
        auto_lr = self._choose_lr(np.array(lrs), np.array(losses_smoothened))

        # log curve and auto_lr to tensorboard as an image
        try:
            # import matplotlib with headless backend
            import matplotlib

            matplotlib.use("Agg")
            import matplotlib.pyplot as plt

            # create AutoLR loss = f(learning_rate) curve
            fig, ax = plt.subplots()
            ax.semilogx(lrs, losses, ".", alpha=0.3, label="Raw loss")
            ax.semilogx(lrs, losses_smoothened, linewidth=2, label="Smoothened loss")
            ax.set_xlabel("Learning rate")
            ax.set_ylabel("Loss")
            ax.legend()

            # indicate selected learning rate by a vertical line
            ax.plot(
                [auto_lr, auto_lr],
                [np.nanmin(losses_smoothened), np.nanmax(losses_smoothened)],
                linewidth=3,
            )

            # zoom on meaningful part of the curve
            m = np.nanmin(losses_smoothened)
            M = 1.1 * losses_smoothened[10]
            ax.set_ylim(m, M)

            # indicate selected learning rate in the figure title
            ax.set_title(f"AutoLR = {auto_lr:g}")

            # send matplotlib figure to Tensorboard
            trainer.tensorboard_.add_figure(
                "train/auto_lr", fig, global_step=trainer.epoch_, close=True
            )

        except ImportError as e:
            msg = (
                f"Something went wrong when trying to send AutoLR figure "
                f"to Tensorboard. Did you install matplotlib?\n\n{e}\n\n"
            )
            print(msg)
            print(e)

        except Exception as e:
            msg = (
                f"Something went wrong when trying to send AutoLR figure "
                f"to Tensorboard. It is OK but you might want to have a "
                f"look at why this happened.\n\n{e}\n\n"
            )
            print(msg)

        return auto_lr


class ConstantScheduler(BaseSchedulerCallback):
    """Constant learning rate"""

    pass


class DavisKingScheduler(BaseSchedulerCallback):
    """Automatic Learning Rate Scheduling That Really Works

    http://blog.dlib.net/2018/02/automatic-learning-rate-scheduling-that.html

    Parameters
    ----------
    factor : float, optional
        Factor by which the learning rate will be reduced.
        new_lr = old_lr * factor. Defaults to 0.5
    patience : int, optional
        Number of epochs with no improvement after which learning rate will
        be reduced. Defaults to 10.
    """

    def __init__(self, factor: float = 0.5, patience: int = 10):
        super().__init__()
        self.factor = factor
        self.patience = patience

    def on_train_start(self, trainer: "Trainer") -> None:
        super().on_train_start(trainer)
        maxlen = 2 * self.patience * trainer.batches_per_epoch
        self.losses_ = deque([], maxlen=maxlen)

    def on_epoch_end(self, trainer: "Trainer") -> None:

        # compute statistics on batch loss trend
        count = steps_without_decrease(np.array(self.losses_))
        count_robust = steps_without_decrease(np.array(self.losses_), robust=True)

        # if batch loss hasn't been decreasing for a while
        patience = self.patience * trainer.batches_per_epoch
        if count > patience and count_robust > patience:
            self.learning_rate = self.factor * self.learning_rate
            self.losses_.clear()

    def on_batch_end(self, trainer, batch_loss):
        super().on_batch_end(trainer, batch_loss)

        # store current batch loss
        self.losses_.append(batch_loss["loss"].item())


class CyclicScheduler(BaseSchedulerCallback):
    """Cyclic learning rate (and momentum)

    Parameters
    ----------
    epochs_per_cycle : int, optional
        Number of epochs per cycle. Defaults to 10.
    decay : {float, 'auto'}, optional
        Update base learning rate at the end of each cycle:
            - when `float`, multiply base learning rate by this amount;
            - when 'auto', apply AutoLR;
            - defaults to doing nothing.

    Reference
    ---------
    Leslie N. Smith. "Cyclical Learning Rates for Training Neural Networks"
    IEEE Winter Conference on Applications of Computer Vision (WACV, 2017).
    """

    def __init__(self, epochs_per_cycle: int = 10, decay=None):
        super().__init__()
        self.epochs_per_cycle = epochs_per_cycle
        self.decay = decay

    @property
    def momentum(self) -> float:
        return self._momentum

    @momentum.setter
    def momentum(self, momentum: float) -> None:
        for g in self.optimizer_.param_groups:
            g["momentum"] = momentum
        self._momentum = momentum

    def on_train_start(self, trainer: "Trainer") -> None:
        """Initialize batch/epoch counters"""

        super().on_train_start(trainer)
        self.batches_per_cycle_ = self.epochs_per_cycle * trainer.batches_per_epoch
        self.n_batches_ = 0
        self.n_epochs_ = 0

        self.learning_rate = trainer.base_learning_rate_ * 0.1

    def on_epoch_end(self, trainer: "Trainer") -> None:
        """Update base learning rate at the end of cycle"""

        super().on_epoch_end(trainer)

        # reached end of cycle?
        self.n_epochs_ += 1
        if self.n_epochs_ % self.epochs_per_cycle == 0:

            # apply AutoLR
            if self.decay == "auto":
                trainer.base_learning_rate_ = self.auto_lr(trainer)

            # decay base learning rate
            elif self.decay is not None:
                trainer.base_learning_rate_ *= self.decay

            # reset epoch/batch counters
            self.n_epochs_ = 0
            self.n_batches_ = 0

    def on_batch_start(self, trainer: "Trainer", batch: dict) -> dict:
        """Update learning rate & momentum according to position in cycle"""

        super().on_batch_start(trainer, batch)

        # position within current cycle (reversed V)
        rho = 1.0 - abs(2 * (self.n_batches_ / self.batches_per_cycle_ - 0.5))

        self.learning_rate = trainer.base_learning_rate_ * (0.1 + 0.9 * rho)
        self.momentum = MOMENTUM_MAX - (MOMENTUM_MAX - MOMENTUM_MIN) * rho

        self.n_batches_ += 1

        return batch



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016-2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import warnings
import numpy as np

from .utils import RawAudio

from pyannote.core import Segment
from pyannote.core import SlidingWindow
from pyannote.core import SlidingWindowFeature

from pyannote.database import get_unique_identifier

from librosa.util import valid_audio
from librosa.util.exceptions import ParameterError


class FeatureExtraction:
    """Base class for feature extraction

    Parameters
    ----------
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    sample_rate : int, optional
        Convert audio to use this sample rate.

    See also
    --------
    `pyannote.audio.augmentation.AddNoise`
    """

    def __init__(self, augmentation=None, sample_rate=None):
        super().__init__()
        self.sample_rate = sample_rate

        # used in FeatureExtraction.crop
        self.raw_audio_ = RawAudio(
            sample_rate=self.sample_rate, mono=True, augmentation=augmentation
        )

    def augmentation():
        doc = "Data augmentation."

        def fget(self):
            return self.raw_audio_.augmentation

        def fset(self, augmentation):
            self.raw_audio_.augmentation = augmentation

        return locals()

    augmentation = property(**augmentation())

    def get_dimension(self):
        """Get dimension of feature vectors

        Returns
        -------
        dimension : int
            Dimension of feature vectors
        """
        msg = "`FeatureExtraction subclasses must implement " "`get_dimension` method."
        raise NotImplementedError(msg)

    @property
    def dimension(self):
        """Dimension of feature vectors"""
        return self.get_dimension()

    def get_resolution(self):
        """Get sliding window used for feature extraction

        Returns
        -------
        sliding_window : `pyannote.core.SlidingWindow`
            Sliding window used for feature extraction.
        """

        msg = (
            "`FeatureExtraction` subclasses must implement " "`get_resolution` method."
        )
        raise NotImplementedError(msg)

    @property
    def sliding_window(self):
        """Sliding window used for feature extraction"""
        return self.get_resolution()

    def get_features(self, y, sample_rate):
        """Extract features from waveform

        Parameters
        ----------
        y : (n_samples, 1) numpy array
            Waveform.
        sample_rate : int
            Sample rate.

        Returns
        -------
        features : (n_frames, dimension) numpy array
            Extracted features
        """
        msg = "`FeatureExtractions subclasses must implement " "`get_features` method."
        raise NotImplementedError(msg)

    def __call__(self, current_file) -> SlidingWindowFeature:
        """Extract features from file

        Parameters
        ----------
        current_file : dict
            `pyannote.database` files.

        Returns
        -------
        features : `pyannote.core.SlidingWindowFeature`
            Extracted features
        """

        # load waveform, re-sample, convert to mono, augment, normalize
        y, sample_rate = self.raw_audio_(current_file, return_sr=True)

        # compute features
        features = self.get_features(y.data, sample_rate)

        # basic quality check
        if np.any(np.isnan(features)):
            uri = get_unique_identifier(current_file)
            msg = f'Features extracted from "{uri}" contain NaNs.'
            warnings.warn(msg.format(uri=uri))

        # wrap features in a `SlidingWindowFeature` instance
        return SlidingWindowFeature(features, self.sliding_window)

    def get_context_duration(self) -> float:
        """

        TODO. explain why this is needed

        Returns
        -------
        context : float
            Context duration, in seconds.
        """
        return 0.0

    def crop(self, current_file, segment, mode="center", fixed=None) -> np.ndarray:
        """Fast version of self(current_file).crop(segment, mode='center',
+                                                  fixed=segment.duration)

        Parameters
        ----------
        current_file : dict
            `pyannote.database` file. Must contain a 'duration' key that
            provides the duration (in seconds) of the audio file.
        segment : `pyannote.core.Segment`
            Segment from which to extract features.

        Returns
        -------
        features : (n_frames, dimension) numpy array
            Extracted features

        See also
        --------
        `pyannote.core.SlidingWindowFeature.crop`
        """

        context = self.get_context_duration()

        # extend segment on both sides with requested context
        xsegment = Segment(
            max(0, segment.start - context),
            min(current_file["duration"], segment.end + context),
        )

        # obtain (augmented) waveform on this extended segment
        y = self.raw_audio_.crop(
            current_file, xsegment, mode="center", fixed=xsegment.duration
        )

        features = self.get_features(y, self.sample_rate)

        # get rid of additional context before returning
        frames = self.sliding_window
        shifted_frames = SlidingWindow(
            start=xsegment.start - frames.step,
            step=frames.step,
            duration=frames.duration,
        )
        ((start, end),) = shifted_frames.crop(
            segment, mode=mode, fixed=fixed, return_ranges=True
        )

        # HACK for when start (returned by shifted_frames.crop) is negative
        # due to floating point precision.
        if start < 0:
            if fixed is not None:
                end -= start
            start = 0

        return features[start:end]



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2017-2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
# Update TristouNet to latest API. Commenting it out for now.
# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO


# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# from torch.nn.utils.rnn import PackedSequence
# from torch.nn.utils.rnn import pad_packed_sequence
#
#
# class TristouNet(nn.Module):
#     """TristouNet sequence embedding
#
#     RNN ( » ... » RNN ) » temporal pooling › ( MLP › ... › ) MLP › normalize
#
#     Parameters
#     ----------
#     specifications : `dict`
#         Batch specifications:
#             {'X': {'dimension': n_features}}
#     rnn : {'LSTM', 'GRU'}, optional
#         Defaults to 'LSTM'.
#     recurrent: list, optional
#         List of output dimension of stacked RNNs.
#         Defaults to [16, ] (i.e. one RNN with output dimension 16)
#     bidirectional : bool, optional
#         Use bidirectional recurrent layers. Defaults to False.
#     pooling : {'sum', 'max'}
#         Temporal pooling strategy. Defaults to 'sum'.
#     linear : list, optional
#         List of hidden dimensions of linear layers. Defaults to [16, 16].
#
#     Reference
#     ---------
#     Hervé Bredin. "TristouNet: Triplet Loss for Speaker Turn Embedding."
#     ICASSP 2017 (https://arxiv.org/abs/1609.04301)
#     """
#
#     supports_packed = True
#
#     def __init__(self, specifications,
#                  rnn='LSTM', recurrent=[16], bidirectional=False,
#                  pooling='sum', linear=[16, 16]):
#
#         super(TristouNet, self).__init__()
#
#         self.specifications = specifications
#         self.n_features_ = specifications['X']['dimension']
#         self.rnn = rnn
#         self.recurrent = recurrent
#         self.bidirectional = bidirectional
#         self.pooling = pooling
#         self.linear = [] if linear is None else linear
#
#         self.num_directions_ = 2 if self.bidirectional else 1
#
#         if self.pooling not in {'sum', 'max'}:
#             raise ValueError('"pooling" must be one of {"sum", "max"}')
#
#         # create list of recurrent layers
#         self.recurrent_layers_ = []
#         input_dim = self.n_features_
#         for i, hidden_dim in enumerate(self.recurrent):
#             if self.rnn == 'LSTM':
#                 recurrent_layer = nn.LSTM(input_dim, hidden_dim,
#                                           bidirectional=self.bidirectional,
#                                           batch_first=True)
#             elif self.rnn == 'GRU':
#                 recurrent_layer = nn.GRU(input_dim, hidden_dim,
#                                          bidirectional=self.bidirectional,
#                                          batch_first=True)
#             else:
#                 raise ValueError('"rnn" must be one of {"LSTM", "GRU"}.')
#             self.add_module('recurrent_{0}'.format(i), recurrent_layer)
#             self.recurrent_layers_.append(recurrent_layer)
#             input_dim = hidden_dim * (2 if self.bidirectional else 1)
#
#         # create list of linear layers
#         self.linear_layers_ = []
#         for i, hidden_dim in enumerate(self.linear):
#             linear_layer = nn.Linear(input_dim, hidden_dim, bias=True)
#             self.add_module('linear_{0}'.format(i), linear_layer)
#             self.linear_layers_.append(linear_layer)
#             input_dim = hidden_dim
#
#     @property
#     def dimension(self):
#         if self.linear:
#             return self.linear[-1]
#         return self.recurrent[-1] * (2 if self.bidirectional else 1)
#
#     def forward(self, sequence):
#         """
#
#         Parameters
#         ----------
#         sequence : (batch_size, n_samples, n_features) torch.Tensor
#
#         """
#
#         packed_sequences = isinstance(sequence, PackedSequence)
#
#         if packed_sequences:
#             _, n_features = sequence.data.size()
#             batch_size = sequence.batch_sizes[0].item()
#             device = sequence.data.device
#         else:
#             # check input feature dimension
#             batch_size, _, n_features = sequence.size()
#             device = sequence.device
#
#         if n_features != self.n_features_:
#             msg = 'Wrong feature dimension. Found {0}, should be {1}'
#             raise ValueError(msg.format(n_features, self.n_features_))
#
#         output = sequence
#
#         # recurrent layers
#         for hidden_dim, layer in zip(self.recurrent, self.recurrent_layers_):
#
#             if self.rnn == 'LSTM':
#                 # initial hidden and cell states
#                 h = torch.zeros(self.num_directions_, batch_size, hidden_dim,
#                                 device=device, requires_grad=False)
#                 c = torch.zeros(self.num_directions_, batch_size, hidden_dim,
#                                 device=device, requires_grad=False)
#                 hidden = (h, c)
#
#             elif self.rnn == 'GRU':
#                 # initial hidden state
#                 hidden = torch.zeros(
#                     self.num_directions_, batch_size, hidden_dim,
#                     device=device, requires_grad=False)
#
#             # apply current recurrent layer and get output sequence
#             output, _ = layer(output, hidden)
#
#         if packed_sequences:
#             output, lengths = pad_packed_sequence(output, batch_first=True)
#
#         # batch_size, n_samples, dimension
#
#         # average temporal pooling
#         if self.pooling == 'sum':
#             output = output.sum(dim=1)
#         elif self.pooling == 'max':
#             if packed_sequences:
#                 msg = ('"max" pooling is not yet implemented '
#                        'for variable length sequences.')
#                 raise NotImplementedError(msg)
#             output, _ = output.max(dim=1)
#
#         # batch_size, dimension
#
#         # stack linear layers
#         for hidden_dim, layer in zip(self.linear, self.linear_layers_):
#
#             # apply current linear layer
#             output = layer(output)
#
#             # apply non-linear activation function
#             output = torch.tanh(output)
#
#         # batch_size, dimension
#
#         # unit-normalize
#         norm = torch.norm(output, 2, 1, keepdim=True)
#         output = output / norm
#
#         return output



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import torch
import torch.nn as nn
from typing import Text
from typing import List
from pyannote.core import SlidingWindow
from pyannote.audio.train.task import Task


class Convolutional(nn.Module):
    """Convolutional layers

    Parameters
    ----------
    n_features : int,
        Input feature shape. Should be 1.
    sample_rate: int, optional
        Input sample rate (in Hz). Defaults to 16000.
    out_channels : list of int, optional
        Number of channels produced by the convolutions.
    kernel_size : list of int, optional
        Size of the convolving kernels.
    stride : list of int, optional
        Stride of the convolutions
    max_pool : list of int, optional
        Size and stride of the size of the windows to take a max over.
    instance_normalize : bool, optional
        Apply instance normalization after pooling. Set to False to not apply
        any normalization. Defaults to True.
    dropout : float, optional
        If non-zero, introduces a Dropout layer on the outputs of each layer
        except the last layer, with dropout probability equal to dropout.
    """

    def __init__(
        self,
        n_features: int,
        sample_rate: int = 16000,
        out_channels: List[int] = [512, 512, 512, 512, 512, 512],
        kernel_size: List[int] = [251, 5, 5, 5, 5, 5],
        stride: List[int] = [5, 1, 1, 1, 1, 1],
        max_pool: List[int] = [3, 3, 3, 3, 3, 3],
        instance_normalize: bool = True,
        dropout: float = 0.0,
    ):
        super().__init__()

        self.n_features = n_features
        self.sample_rate = sample_rate
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.max_pool = max_pool
        self.instance_normalize = instance_normalize
        self.dropout = dropout

        self.conv1ds = nn.ModuleList()
        self.pool1ds = nn.ModuleList()
        if self.instance_normalize:
            self.norm1ds = nn.ModuleList()
        self.activation = nn.LeakyReLU(negative_slope=1e-2, inplace=False)
        if self.dropout > 0.0:
            self._dropout = nn.Dropout(p=self.dropout, inplace=False)

        in_channels = self.n_features
        for i, (out_channels, kernel_size, stride, max_pool) in enumerate(
            zip(self.out_channels, self.kernel_size, self.stride, self.max_pool)
        ):
            conv1d = nn.Conv1d(
                in_channels,
                out_channels,
                kernel_size,
                stride=stride,
                padding=0,
                dilation=1,
                groups=1,
                bias=True,
                padding_mode="zeros",
            )
            self.conv1ds.append(conv1d)

            pool1d = nn.MaxPool1d(
                max_pool,
                stride=max_pool,
                padding=0,
                dilation=1,
                return_indices=False,
                ceil_mode=False,
            )
            self.pool1ds.append(pool1d)

            if self.instance_normalize:
                norm1d = nn.InstanceNorm1d(out_channels, affine=True)
                self.norm1ds.append(norm1d)

            in_channels = out_channels

    def forward(self, waveforms: torch.Tensor) -> torch.Tensor:
        """Forward pass

        Parameter
        ---------
        waveform : torch.Tensor
            Waveform tensor with shape (batch_size, n_samples, 1)

        Returns
        -------
        features : torch.Tensor
            Output features (batch_size, n_frames, n_features)
        """

        output = waveforms.transpose(1, 2)

        for i, (conv1d, pool1d) in enumerate(zip(self.conv1ds, self.pool1ds)):
            output = conv1d(output)
            output = pool1d(output)
            if self.instance_normalize:
                output = self.norm1ds[i](output)
            output = self.activation(output)
            if self.dropout > 0.0 and i + 1 < self.num_layers:
                output = self._dropout(output)

        return output.transpose(1, 2)

    @property
    def num_layers(self):
        """Number of convolutional layers"""
        return len(self.out_channels)

    @property
    def dimension(self):
        """Dimension of output features"""
        return self.out_channels[-1]

    @staticmethod
    def get_alignment(task: Task, **kwargs) -> Text:
        """Output frame alignment"""
        return "strict"

    @staticmethod
    def get_resolution(
        task: Task,
        sample_rate: int = 16000,
        out_channels: List[int] = [512, 512, 512, 512, 512, 512],
        kernel_size: List[int] = [251, 5, 5, 5, 5, 5],
        stride: List[int] = [5, 1, 1, 1, 1, 1],
        max_pool: List[int] = [3, 3, 3, 3, 3, 3],
        **kwargs,
    ) -> SlidingWindow:
        """Output frame resolution"""

        # https://medium.com/mlreview/a-guide-to-receptive-field-arithmetic-for-convolutional-neural-networks-e0f514068807
        padding = 0
        receptive_field, jump, start = 1, 1, 0.5
        for ks, s, mp in zip(kernel_size, stride, max_pool):
            # increase due to (Sinc)Conv1d
            receptive_field += (ks - 1) * jump
            start += ((ks - 1) / 2 - padding) * jump
            jump *= s
            # increase in receptive field due to MaxPool1d
            receptive_field += (mp - 1) * jump
            start += ((mp - 1) / 2 - padding) * jump
            jump *= mp

        return SlidingWindow(
            duration=receptive_field / sample_rate, step=jump / sample_rate, start=0.0
        )
